<?php
/*
This file is part of Formol.

Formol is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

Formol is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * Field class
 *
 * Manages field behaviour
 *
 * @copyright Copyright (c) 2007 Thomas Mazière
 * @author Thomas Mazière <thomas.maziere@gmail.com>
 * @package formol
 */
require_once('Element.class.php');

//Renderers
define('RENDERER_PATH', dirname(__FILE__).'/Renderers/');
require_once(RENDERER_PATH.'InputTextRenderer.class.php');
require_once(RENDERER_PATH.'InputPasswordRenderer.class.php');
require_once(RENDERER_PATH.'InputHiddenRenderer.class.php');
require_once(RENDERER_PATH.'SelectRenderer.class.php');
require_once(RENDERER_PATH.'TextareaRenderer.class.php');
require_once(RENDERER_PATH.'InputRadioRenderer.class.php');
require_once(RENDERER_PATH.'InputCheckboxRenderer.class.php');
require_once(RENDERER_PATH.'InputSubmitRenderer.class.php');
require_once(RENDERER_PATH.'InputImageRenderer.class.php');
require_once(RENDERER_PATH.'InputResetRenderer.class.php');
require_once(RENDERER_PATH.'InputButtonRenderer.class.php');
require_once(RENDERER_PATH.'InputFileRenderer.class.php');
//Validators
define('VALIDATOR_PATH', dirname(__FILE__).'/Validators/');
require_once(VALIDATOR_PATH.'RequiredValidator.class.php');
require_once(VALIDATOR_PATH.'MaxValidator.class.php');
require_once(VALIDATOR_PATH.'MinValidator.class.php');
require_once(VALIDATOR_PATH.'EqualsValidator.class.php');
require_once(VALIDATOR_PATH.'ExpregValidator.class.php');
require_once(VALIDATOR_PATH.'EmailValidator.class.php');
require_once(VALIDATOR_PATH.'TelValidator.class.php');

class Field extends Element
{

	/**
	 * Field name
	 *
	 * @var string
	 */
	protected $_name;

	/**
	 * Field type
	 *
	 * @var string
	 */
	protected $_type;

	/**
	 * Field label
	 *
	 * @var string
	 */
	protected $_label;

	/**
	 * Field default value
	 *
	 * @var string
	 */
	protected $_defValue;

	/**
	 * Field current value
	 *
	 * @var string
	 */
	protected $_curValue;

	/**
	 * Css class name for label
	 *
	 * @var string
	 */
	protected $_labelClass;

	/**
	 * Css class name for field body
	 *
	 * @var string
	 */
	protected $_fieldClass;

	/**
	 * field enabled ?
	 *
	 * @var bool
	 */
	protected $_enabled = true;

	/**
	 * field to be saved ?
	 *
	 * @var bool
	 */
	protected $_saved = true;
	

	/**
	 * Array of registered validators
	 *
	 * @var array
	 */
	protected $_validators = array();

	/**
	 * Array of registered preProcessors
	 * (used before validation of a new value)
	 *
	 * @var array
	 */
	protected $_preProcessors = array();

	/**
	 * Array of registered posProcessors
	 * (used after validation of a new value)
	 *
	 * @var array
	 */
	protected $_postProcessors = array();

	/**
	 * Object which renders current field
	 *
	 * @var FieldRenderer
	 */
	public $renderer;
	
	/**
	 * errors
	 *
	 * @var array
	 */
	public $errors = array();


	/**
	 * Field types const
	 *
	 * @var string
	 */
	const TEXT = "InputText";
	const SELECT = "Select";
	const RADIO = "InputRadio";
	const CHECKBOX = "InputCheckbox";
	const PASSWORD = "InputPassword";
	const TEXTAREA = "Textarea";
	const HIDDEN = "InputHidden";
	const BUTTON = "InputButton";
	const SUBMIT = "InputSubmit";
	const RESET = "InputReset";
	const IMAGE = "InputImage";
	const FILE = "InputFile";
	
	
	
	/**
	 * Constructor
	 *
	 * @param string $name
	 * @param string $type	 
	 * @return bool
	 */

	public function __construct($name, $type) {
		$type = strtoupper($type);
		$name = trim($name);
		if (defined('self::'.$type)) {
			$this->_name = $name;
			$this->_type = constant('self::'.$type);
			$renderer = constant('self::'.$type).'Renderer';
			$this->renderer = new $renderer($this);
		} else {		
			throw new FormException("Field type unknown");
		}
	}

	/**
	 * Returns property if allowed
	 *
	 * @return mixed
	 */

	public function __get($prop) {
		$allow = array(	'name'=>'name',
						'type'=>'type',
						'defValue'=>'defValue',
						'value'=>'curValue',
						'labelClass'=>'labelClass',
						'fieldClass'=>'fieldClass',
						'label'=>'label',
						'enabled'=>'enabled',
						'saved'=>'saved',
						'preProcessors'=>'preProcessors',
						'validators'=>'validators',
						'postProcessors'=>'postProcessors'
						);
		if (array_key_exists($prop, $allow)) {
			$prop = '_'.$allow[$prop];
			return $this->$prop;
		} else {
			throw new FormException("Unknown or forbidden property '".htmlentities($prop)."'", 1);
		}
	}
	
	/**
	 * Sets property if allowed
	 *
	 * @return object Field
	 */

	public function __set($prop, $value) {
		$allow = array(	'default'=>'defValue',
						'label'=>'label',
						'defValue'=>'defValue',
						'value'=>'curValue',
						'labelClass'=>'labelClass',
						'fieldClass'=>'fieldClass',
						'enabled'=>'enabled',
						'saved'=>'saved',
						'errors'=>'errors',
						'xhtmlPattern'=>'xhtmlPattern'
						);
		if (array_key_exists($prop, $allow)) {
			return $this->setProperty($allow[$prop], $value);
		} else {
			throw new FormException("Unknown or forbidden property '".htmlentities($prop)."' (access by property)");
		}
	}
	
	/**
	 * Sets property if allowed (through method call)
	 *
	 * @return object Field
	 */
	
	public function __call($function, $args) {
		$allow = array(	'default'=>'defValue',
						'label'=>'label',
						'defValue'=>'defValue',
						'value'=>'curValue',
						'labelClass'=>'labelClass',
						'fieldClass'=>'fieldClass',
						'enabled'=>'enabled',
						'saved'=>'saved',
						'errors'=>'errors',
						'xhtmlPattern'=>'xhtmlPattern'						
						);
		if (array_key_exists($function, $allow)) {
			return $this->setProperty($allow[$function], $args[0]);
		} else {
			throw new FormException("Unknown or forbidden property '".htmlentities($function)."' (access by method)");
		}
    }
	
	/**
	 * Sets property
	 *
	 * @return object Field
	 */

	private function setProperty($prop, $value) {
		$prop = '_'.$prop;
		switch ($prop) {
			case '_defValue':
				$this->_defValue = $value;
				$this->_curValue = $value;
			break;
			case '_xhtmlPattern':
				$this->renderer->setXhtmlPattern($value);
			break;
			default:
				$this->$prop = $value;
			break;
		}	
		return $this;
	}

	/**
	 * Renders field
	 *
	 * @return string
	 */

	public function __toString() {
		return $this->render();
	}


	/**
	 * Renders whole field
	 *
	 * @return string
	 */

	public function render($args = array()) {
		return $this->renderer->render($args);
	}
	
	/**
	 * Renders field label
	 *
	 * @return string
	 */

	public function renderLabel() {
		return $this->renderer->renderLabel();
	}
	
	/**
	 * Renders field body
	 *
	 * @return string
	 */

	public function renderField($args = array()) {
		return $this->renderer->renderField($args);
	}


	/**
	 * Process field value
	 *
	 */

	public function process() {

	}


	/**
	 * Validate field value
	 *
	 */

	public function validate() {

	}


	/**
	 * register field preProcessor
	 *
	 * @return void
	 */

	public function registerPreProcess($process) {
		$process = htmlentities(strtolower($process));
		if (!in_array($process, $this->_preProcessors)) {
			$this->_preProcessors[] = $process;
			return $this;
		} else {
			throw new FormException("PreProcess '".$process."' already registered");
		}
	}
	
	/**
	 * register field postProcessor
	 *
	 * @return void
	 */

	public function registerPostProcess($process) {
		$process = htmlentities(strtolower($process));
		if (!in_array($process, $this->_postProcessors)) {
			$this->_postProcessors[] = $process;
			return $this;
		} else {
			throw new FormException("PostProcess '".$process."' already registered");
		}
	}
	
	/**
	 * register field Validator
	 *
	 * @return void
	 */

	public function registerValidator($validator, $errorMessage = '') {
		$validator = ucfirst(strtolower($validator));
		$args = array();
		if (strpos($validator, ':') !== false) {
			$elems = explode(':', $validator);
			$validator = $elems[0];
			$args = explode(',', $elems[1]);
		}
		$errorMessage = trim($errorMessage);
		if (class_exists($validator.'Validator')) {
			if (!in_array($validator, $this->_validators)) {
				$this->_validators[] = array($validator, $errorMessage, $args);
				return $this;
			} else {
			throw new FormException("Validator '".$validator."' already registered");
			}
		} else {
			throw new FormException("Unknown validator '".$validator."'");
		}
	}	
	
	
	/**
	 * export to xml
	 *
	 * @return string
	 */

	public function toXML() {
		$strProperties = array('name', 'type', 'label', 'defValue', 'labelClass', 'fieldClass', 'enabled', 'saved');
		$arrayProperties = array('preProcessors', 'postProcessors');
		$xml .= "\t<field";
		// String properties
		foreach ($strProperties as $prop) {
			$value = ($this->$prop != '') ? $this->$prop : " ";
			$xml .= " $prop=\"$value\"";
		}
		$xml .= ">\n";
		// Array properties
		foreach ($arrayProperties as $prop) {
			if (count($this->$prop) > 0) {
				$xml .= "\t\t<".$prop.">\n";
				foreach ($this->$prop as $elem) {
					$xml .= "\t\t\t<".$elem." />\n";
				}
				$xml .= "\t\t</".$prop.">\n";
			} else {
				$xml .= "\t\t<".$prop." />\n";
			}
		}
		// Validators
		if (count($this->validators) > 0) {
			$xml .= "\t\t<validators>\n";		
			foreach ($this->validators as $validator) {
				$xml .= "\t\t\t<".$validator[0];
				$message = ($validator[1] != '') ? $validator[1] : "";
				$xml .= " errorMessage=\"".$message."\" />\n";
			}
			$xml .= "\t\t</validators>\n";
		} else {
			$xml .= "\t\t<validators />\n";	
		}
		$xml .= "\t</field>\n";
		return $xml;
	}
	
}
?>
