<?php
/*
This file is part of Formol.

Formol is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

Formol is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
	
/**
 * Form class
 *
 * Manages fields sequence and consistency
 *
 * @copyright Copyright (c) 2007 Thomas Mazière
 * @author Thomas Mazière <thomas.maziere@gmail.com>
 * @package formol
 */
require_once('FormException.class.php');
require_once('Field.class.php');
require_once('FormManager.class.php');
require_once('Element.class.php');


class Form implements Iterator
{

	/**
	 * Array of fields
	 * @var array
	 */
	private $_fields = array();

	/**
	 * Form id
	 * @var int
	 */
	private $_id;

	/**
	 * Form instance
	 * @var object
	 * @static
	 */
	private static $instance = null;
	
	/**
	 * Form manager
	 * @var object FormManager
	 */
	private $_manager;
	
	/**
	 * Form action
	 * @var string
	 */
	private $_action;
	
	/**
	 * Form enctype
	 * @var string
	 */
	private $_enctype;
	
	/**
	 * Array of Errors
	 * @var array
	 */
	public $errors = array();

	
	/**
	 * Constructor
	 *
	 */

	private function __construct() {}
	
	/**
	 * create new unique instance.
	 * erase existing one
	 *
	 * @return object
	 * @static
	 */
	 
	public static function create() {
		if (!is_null(self::$instance)) {
			self::$instance = null;
		}
		self::$instance = new Form();
		self::$instance->_id = md5( uniqid() );
		self::$instance->_manager = new FormManager(self::$instance);
		return self::$instance;
	}
	
	/**
	 * return unique form instance (singleton)
	 * creates one if necessary
	 * @return object
	 * @static
	 */

	public static function getInstance() {
		if (is_null(self::$instance)) {
			return self::create();
		} else {
			return self::$instance;
		}
	}
	
	
	/**
	 * Save form in session or other persistence manager
	 *
	 * @static
	 * @param mixed $persistenceManager array (ex. $_SESSION) or object
	 * @return bool
	 */

	public static function save(&$persistenceManager) {
		if (is_array($persistenceManager)) {
			$persistenceManager['form'] = self::$instance;
			return true;
		} else if (is_object($persistenceManager)) {
			$persistenceManager->form = self::$instance;
			return true;
		} else {
			throw new FormException("Wrong parameter type : must be array or object");
		}
	}
	
	/**
	 * Load form from session or other persistence manager
	 *
	 * @static	 
	 * @param mixed $persistenceManager array (ex. $_SESSION) or object
	 * @return mixed Form instance or false
	 */

	public static function load(&$persistenceManager, $postVars) {
		if (is_array($persistenceManager) && isset($persistenceManager['form'])  && ($persistenceManager['form'] instanceof Form)) {
			self::$instance = $persistenceManager['form'];	
			if (self::$instance->update($postVars)) {
				return self::getInstance();
			} else {
				return false;
			}
		} else if (is_object($persistenceManager) && isset($persistenceManager->form) && ($persistenceManager->form instanceof Form)) {
			self::$instance = $persistenceManager->form;
			if (self::$instance->update($postVars)) {
				return self::getInstance();
			} else {
				return false;
			}
		} else {
			return false;
		}
	}
	

	/**
	 * Free form instance
	 *
	 * @return bool
	 */

	public static function free(&$persistenceManager = null) {
		if (!is_null($persistenceManager)) {
			if (is_array($persistenceManager) && isset($persistenceManager['form'])  && ($persistenceManager['form'] instanceof Form)) {
				unset($persistenceManager['form']);
			} else if (is_object($persistenceManager) && isset($persistenceManager->form) && ($persistenceManager->form instanceof Form)) {
				$persistenceManager->form = null;
			} else {
				throw new FormException("Unable to free Form instance");
			}
		}
		self::$instance = null;
		return true;
	}
	

	/**
	 * Updates the submitted form
	 *
	 * Looks for $postVars['submitted'], then $postVars[$this->_id]
	 * then calls FormUpdater
	 * @param array $postVars
	 * @return bool
	 */

	private function update($postVars) {
		if (is_array($postVars) && array_key_exists('submitted', $postVars)) {
			if (array_key_exists('formId', $postVars) && $postVars['formId'] == $this->_id) {
				$this->_manager->update($postVars);
				return true;
			} else {
				self::free();
				return false;
			}
		} else {
			self::free();
			return false;
		}
	}
	
	/**
	 * Validates the form
	 *
	 * checks if any error registered for any field
	 *
	 * @return bool
	 */

	public function validate() {
		return $this->_manager->validate();
	}

	
	/**
	 * Returns field object (by name)
	 *
	 * @param string $fieldName
	 * @return Field object
	 */

	public function __get($name) {
		
		if (property_exists('Form', '_'.$name)) {
			//echo "<br />==> "; echo $prop = '_'.$name; echo "<br />";
			var_dump($this->$prop);
			return $this->$prop;
		} else if (array_key_exists($name, $this->_fields)) {
			return $this->_fields[$name];
		} else {
			throw new FormException("Unknown form property (".htmlentities($name).")");
		}
	}
	
	/**
	 * Sets property if allowed
	 *
	 * @return object Form
	 */

	public function __set($prop, $value) {
		$allow = array(	'action'=>'action');
		if (array_key_exists($prop, $allow)) {
			return $this->setProperty($allow[$prop], $value);
		} else {
			throw new FormException("Unknown or forbidden property '".htmlentities($prop)."' (access by property)");
		}
	}
	
	/**
	 * Sets property
	 *
	 * @return object Field
	 */

	private function setProperty($prop, $value) {
		$prop = '_'.$prop;
		switch ($prop) {
			default:
				$this->$prop = $value;
			break;
		}	
		return $this;
	}

	/**
	 * Register event for all fields of defined type
	 *
	 * @return object form
	 */

	public function register($event, $value, $type, $arg='') {
		$allow = array('preProcess', 'postProcess', 'validator');
		if (in_array($event, $allow)) {
			$method = 'register'.ucfirst($event);
			foreach ($this as $field) {
				if ($field->type == $type) {
					if (trim($arg) != '') {
						$field->$method($value, $arg);
					} else {
						$field->$method($value);
					}
				}
			}
			return $this;
		} else {
			throw new FormException("Cannot register '".htmlentities($event)."'");
		}		
	}
	
	/**
	 * Sets property for all fields of defined type
	 *
	 * @return object form
	 */

	public function set($prop, $value, $type) {
		foreach ($this as $field) {
			if ($field->type == $type) {
				$field->$prop = $value;
			}
		}
		return $this;
	}

	/**
	 * Creates new field
	 *
	 * @return object Field
	 */

	public function add($name, $type) {
		if(!preg_match('`^[[:alnum:]]{1,30}$`', $name)) { 
			throw new FormException("Cannot add field '".htmlentities($name)."' (rule: 1-30 chars, alphanum)");
		}	
		if (!property_exists('Form', '_'.$name) && !array_key_exists($name, $this->_fields)) {
			$this->_fields[$name] = new Field($name, $type);
			if (strtolower($type) == 'file') {
				$this->_enctype = 'multipart/form-data';
			}
			return $this->_fields[$name];
		} else {
			throw new FormException("Cannot add field '".htmlentities($name)."' : field name already used");
		}
	}
	
	/**
	 * Display form header
	 *
	 *@param string $target
	 * @return string
	 */
	 
	public function header($target=null) {
		$action = '';
		if (!is_null($this->_action)) {
			$action = $this->_action;
		} else if (!is_null($target)) {
			$action = $target;
		} else {
			$action = "";
		}
		$enctype = (trim($this->_enctype) != '') ? " enctype=\"".$this->_enctype."\"" : "";
		$xhtml = "<form action=\"".$action."\" method=\"post\" id=\"".$this->_id."\"".$enctype.">\n";
		$xhtml .= "<input type=\"hidden\" name=\"submitted\" value=\"true\" />\n";
		$xhtml .= "<input type=\"hidden\" name=\"formId\" value=\"".$this->_id."\" />\n";
		return $xhtml;
	}
	
	/**
	 * Display form footer
	 *
	 * @return string
	 */
	 
	public function footer() {
		$xhtml = "</form>\n";
		return $xhtml;
	}
	
	/**
	 * Display the whole form
	 *
	 * @return string
	 */
	 
	public function __toString() {
		return '<h2>Formulaire #'.$this->_id.'</h2>';
	}
	
	/**
	 * Returns current element (HTML
	 *
	 * @return string
	 */

	public function current() {
		return current($this->_fields);
	}

	/**
	 * Returns current field
	 *
	 * @return string
	 */

	public final function key() {
		return key($this->_fields);
	}


	/**
	 * Sets cursor to next field
	 *
	 * @return string
	 */

	public final function next() {
		return next($this->_fields);
	}


	/**
	 * Set fields array to first field
	 *
	 * @return string
	 */

	public final function rewind() {
		reset($this->_fields);
	}


	/**
	 * Returns true if not end of fields array
	 *
	 * @return string
	 */

	public final function valid() {
		$var = $this->current() !== false;
        return $var;
	}
	
	/**
	 * Saves form structure and properties to file (XML)
	 *
	 * @return string
	 */

	public function saveToFile($file) {
		$xml = "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n<form>\n";
		foreach ($this as $field) {
			$xml .= $field->toXML();
		}
		$xml .= "</form>";
		if (is_dir(dirname($file))) {
			$f = fopen($file, 'w');
			fwrite($f, $xml);
			fclose($f);
			return true;
		} else {
			throw new FormException("Cannot save form to xml : ".dirname($file)." is not a valid path");
		}
	}
	
	/**
	 * Loads form from file (XML)
	 *
	 * creates new instance with loaded properties
	 * @return object Form
	 */

	public static function loadFromFile($file) {
		if (is_file($file)) {
			$sxml = simplexml_load_file($file);
			$instance = self::create();
			foreach ($sxml->field as $field) {
				if (isset($field['name']) && isset($field['type'])) {
					$newfield = $instance->add((string) $field['name'], (string) $field['type']);
				} else {
					throw new FormException("Unable to load form : incorrect field data");
				}
				foreach ($field->attributes() as $attr => $val) {
					if ($attr != 'name' && $attr != 'type') {
						$newfield->$attr((string) $val);
					}
				}
				foreach ($field->preProcessors->children() as $process => $value) {
					$newfield->registerPreProcess($process);
				}
				foreach ($field->postProcessors->children() as $process => $value) {
					$newfield->registerPostProcess($process);
				}
				foreach ($field->validators->children() as $validator => $value) {
					$newfield->registerValidator($validator, (string) $value['errorMessage']);
				}
			}
			return $instance;
		} else {
			throw new FormException("Unable to load form : ".$file." is not a valid file");
		}
	}

}
?>
